/*
 * $Id: CustomValidationDemo.cs,v 1.2 2009/01/29 21:32:19 bkline Exp bkline $
 *
 * Proof-of-concept for custom validation in the CDR using .NET.
 *
 * $Log: CustomValidationDemo.cs,v $
 * Revision 1.2  2009/01/29 21:32:19  bkline
 * Added code to populate cdr-eid attributes.
 *
 * Revision 1.1  2009/01/29 20:02:52  bkline
 * Initial revision
 */
using System;
using System.Collections.Generic;
using System.Data;
using System.Data.SqlClient;
using System.IO;
using System.Xml;
using System.Xml.Linq;
using System.Xml.XPath;
using System.Xml.Xsl;

namespace Cdr {

    /*
     * For right now, all we need this class for is the namespace property.
     */
    public class Schema {
        private static XNamespace ns = "http://www.w3.org/2001/XMLSchema";
        public static XNamespace NS { get { return ns; } }
    }

    /*
     * For right now, all we need this class for is the namespace property.
     */
    public class Xslt {
        private static XNamespace ns = "http://www.w3.org/1999/XSL/Transform";
        public static XNamespace NS { get { return ns; } }
    }

    /*
     * An object representing the location and description of an
     * error found while validating a CDR document.  Optionally
     * has an error ID, if error ID attributes are present in the
     * validated document.
     */
    public class ValidationError {
        private string id;
        private string location;
        private string description;
        public ValidationError(XElement e) {
            var options = StringSplitOptions.None;
            var pieces = e.Value.Split(new char[] {':'}, 2, options);
            location = pieces[0];
            description = pieces[1];
            id = (string)e.Attribute("cdr-eid");
        }
        public string ID { get { return id; } }
        public string Location { get { return location; } }
        public string Description { get { return description; } }
        public override string ToString() {
            return String.Format("eid: {0}\npath: {1}\ndescription: {2}\n",
                    id, location, description);
        }
    }

    /*
     * An object representing a condition the users want to have
     * tested at validation time, and for which XML Schemas are
     * not able to check, along with an error message to be
     * displayed when the assertion is broken.
     */
    public class RuleAssertion {
        private string test;
        private string message;
        public RuleAssertion(XElement e) {
            test = (string)e.Attribute("test");
            message = e.Value;
        }
        public XElement toXslt() {
            return new XElement(Xslt.NS + "if",
                new XAttribute("test", "not(" + test + ")"),
                new XElement(Xslt.NS + "call-template",
                    new XAttribute("name", "packError"),
                    new XElement(Xslt.NS + "with-param",
                        new XAttribute("name", "msg"),
                        new XAttribute("select", quoteMessage()))));
        }
        private string quoteMessage() {
            return "'" + message.Replace("'", "&#39;") + "'";
        }
    }

    /*
     * There's supposed to be one of these for every unique path
     * where custom rules should be invoked, with one or more
     * assertion specifying what the users expect to be true
     * at that location in the document.
     */
    public class CustomRule {
        private string context;
        private List<RuleAssertion> assertions = new List<RuleAssertion>();
        public CustomRule(XElement e) {
            context = (string)e.Attribute("context");
            foreach (XElement c in e.Elements(Schema.NS + "assert")) {
                assertions.Add(new RuleAssertion(c));
            }
        }
        public XElement toXslt() {
            var template = new XElement(Xslt.NS + "template");
            template.Add(new XAttribute("match", context));
            foreach (var assertion in assertions)
                template.Add(assertion.toXslt());
            template.Add(new XElement(Xslt.NS + "apply-templates"));
            return template;
        }
    }

    /*
     * Re-implementation of the support for custom validation rules
     * in the C++ CdrXsd module.  Collects all of the custom rules
     * from a CDR schema document.  Doesn't process XSL/T include
     * statements, but that's not hard to add.  The only things
     * exposed are the construction and the method to used the
     * collected rules to validate a CDR document.
     */
    public class CustomRuleSet {
        private List<CustomRule> rules;
        public CustomRuleSet(Doc schema) {
            rules = new List<CustomRule>();
            var root = XElement.Parse(schema.DocXml);
            var xpath = "xsd:annotation/xsd:appInfo/xsd:pattern/xsd:rule";
            var nsManager = new XmlNamespaceManager(new NameTable());
            nsManager.AddNamespace("xsd", Schema.NS.NamespaceName);
            foreach (var e in root.XPathSelectElements(xpath, nsManager))
                rules.Add(new CustomRule(e));
        }

        /*
         * This is the payoff: a list of validation errors found
         * for violation of the custom rules.
         */
        public List<ValidationError> apply(Doc doc) {
            var list = new List<ValidationError>();
            var errors = new XDocument();
            var script = toXslt();
            //Console.WriteLine("xslt doc:\n{0}", script);
            using (var writer = errors.CreateWriter()) {
                var xsl = new XslCompiledTransform();
                xsl.Load(script.CreateReader());
                xsl.Transform(doc.ReaderWithEids, writer);
            }
            //Console.WriteLine("errorDoc:\n{0}", errors);
            foreach (var e in errors.XPathSelectElements("Errors/Err"))
                list.Add(new ValidationError(e));
            return list;
        }
        private XDocument toXslt() {
            var root = new XElement(Xslt.NS + "transform");
            root.Add(new XAttribute("version", "1.0"));
            root.Add(new XAttribute(XNamespace.Xmlns + "cdr", Doc.NS));
            root.Add(makeRootTemplate());
            root.Add(makeRecursionTemplate());
            root.Add(makePackErrorTemplate());
            root.Add(makeGetPathTemplate());
            foreach (var rule in rules)
                root.Add(rule.toXslt());
            return new XDocument(root);
        }
        private static XElement makeRootTemplate() {
            return new XElement(Xslt.NS + "template",
                new XAttribute("match", "/"),
                new XElement("Errors",
                    new XElement(Xslt.NS + "apply-templates")));
        }
        private static XElement makeRecursionTemplate() {
            return new XElement(Xslt.NS + "template",
                new XAttribute("match", "node()"),
                new XElement(Xslt.NS + "apply-templates"));
        }
        private static XElement makePackErrorTemplate() {
            return new XElement(Xslt.NS + "template",
                new XAttribute("name", "packError"),
                new XElement(Xslt.NS + "param",
                    new XAttribute("name", "msg")),
                new XElement(Xslt.NS + "element",
                    new XAttribute("name", "Err"),
                    new XElement(Xslt.NS + "if",
                        new XAttribute("test", "./@cdr-eid"),
                        new XElement(Xslt.NS + "attribute",
                            new XAttribute("name", "cdr-eid"),
                            new XElement(Xslt.NS + "value-of",
                                new XAttribute("select", "./@cdr-eid")))),
                    new XElement(Xslt.NS + "call-template",
                        new XAttribute("name", "getPath"),
                        new XElement(Xslt.NS + "with-param",
                            new XAttribute("name", "nodePart"),
                            new XAttribute("select", "."))),
                    new XElement(Xslt.NS + "value-of",
                        new XAttribute("select", "concat(':', $msg)"))));
        }
        private static XElement makeGetPathTemplate() {
            var select1 = "count($nodePart/preceding-sibling::*) + 1";
            var select2 = "concat('/',$myName,'[',$myPos,']',$childPart)";
            var select3 = "concat('/',name($nodePart),$childPart)";
            return new XElement(Xslt.NS + "template",
                new XAttribute("name", "getPath"),
                new XElement(Xslt.NS + "param",
                    new XAttribute("name", "childPart")),
                new XElement(Xslt.NS + "param",
                    new XAttribute("name", "nodePart")),
                new XElement(Xslt.NS + "variable",
                    new XAttribute("name", "parent"),
                    new XAttribute("select", "$nodePart/parent::node()")),
                new XElement(Xslt.NS + "variable",
                    new XAttribute("name", "myName"),
                    new XAttribute("select", "name($nodePart)")),
                new XElement(Xslt.NS + "choose",
                    new XElement(Xslt.NS + "when",
                        new XAttribute("test", "name($parent)"),
                        new XElement(Xslt.NS + "variable",
                            new XAttribute("name", "myPos"),
                            new XAttribute("select", select1)),
                        new XElement(Xslt.NS + "call-template",
                            new XAttribute("name", "getPath"),
                            new XElement(Xslt.NS + "with-param",
                                new XAttribute("name", "childPart"),
                                new XAttribute("select", select2)),
                            new XElement(Xslt.NS + "with-param",
                                new XAttribute("name", "nodePart"),
                                new XAttribute("select", "$parent")))),
                    new XElement(Xslt.NS + "otherwise",
                        new XElement(Xslt.NS + "value-of",
                            new XAttribute("select", select3)))));
        }
    }

    /*
     * Provides a factory method for creating a connection to the
     * CDR database.  Connection pooling is handled by .NET.  Caller
     * is responsible for closing the connection.
     */
    public class Database {
        public static SqlConnection connect(string uid, string host) {
            string pwd = "eggro11";
            switch (uid) {
            case "CdrGuest":
                pwd = "readonly";
                break;
            case "CdrPublishing":
                pwd = "pubs4Cdr@CIPS";
                break;
            case "cdr":
                break;
            default:
                string err = "unrecognized account: " + uid;
                throw new Exception(err);
            }
            string dsn = String.Format("Database=cdr;User ID={0};" +
                                       "Password={1};" +
                                       "Server={2}", uid, pwd, host);
            SqlConnection c = new SqlConnection(dsn);
            c.Open();
            return c;
        }
    }

    /*
     * Object representing a CDR document retrieved from the database.
     * Very crude version, only dealing with the current working
     * document.  Exposes a property providing an XmlReader object
     * so we can feed the document to a parser or a transformation
     * filter.
     */
    public class Doc {
        private static XNamespace ns = "cips.nci.nih.gov/cdr";
        private int id = 0;
        private string docXml = null;
        public Doc(SqlConnection conn, int docId) {
            id = docId;
            string sql = "SELECT xml FROM document WHERE id = @id";
            SqlCommand cmd = new SqlCommand(sql, conn);
            cmd.Parameters.Add("@id", SqlDbType.Int);
            cmd.Parameters[0].Value = docId;
            SqlDataReader reader = cmd.ExecuteReader();
            if (!reader.Read()) {
                reader.Close();
                string err = String.Format("CDR{0} not found", docId);
                throw new Exception(err);
            }
            docXml = (string)reader[0];
            reader.Close();
        }
        public XmlReader ReaderWithEids { 
            get { 
                var root = XElement.Parse(docXml);
                var start = DateTime.Now;
                Int32 eid = 0;
                foreach (var e in root.DescendantsAndSelf())
                    if (e.NodeType == XmlNodeType.Element)
                        e.SetAttributeValue("cdr-eid", "_" +
                                Convert.ToString(++eid));
                var elapsed = DateTime.Now.Subtract(start);
                Console.WriteLine("populated {0} eids in {1} ms.", eid, 
                        elapsed.TotalMilliseconds);
                return root.CreateReader();
            }
        }
        public XmlReader Reader { 
            get { 
                return XmlReader.Create(new StringReader(docXml)); 
            }
        }
        public int ID { get { return id; } }
        public string DocXml { get { return docXml; } }
        public static XNamespace NS { get { return ns; } }
    }
    /*
     * Throwaway test driver.
     */
    public class CustomValidationDemo {
        public static void Main(string[] args) {
            int docId = Int32.Parse(args[0]);
            SqlConnection conn = Database.connect("cdr", "mahler.nci.nih.gov");
            Doc schema = new Doc(conn, 198); // Common Protocol Schema
            Doc doc = new Doc(conn, docId);
            CustomRuleSet ruleSet = new CustomRuleSet(schema);
            foreach (var error in ruleSet.apply(doc))
                Console.WriteLine(error);
            conn.Close();
        }
    }
}
